package dev.pfaff.altargui.async;

import dev.pfaff.altargui.util.TaskException;

import java.util.concurrent.Future;

/**
 * @param <V> the value produced by the task upon completion.
 */
public interface Task<V> {
	V join() throws TaskException, InterruptedException;

	V joinUninterruptibly() throws TaskException;

	/**
	 * If this method returns {@code true}, then {@link #join()} and friends are guaranteed not to block.
	 */
	boolean isCompleted();

	interface Cancellable<V> extends Task<V> {
		void cancel();
	}

	static <V> Task<V> completedTask(V value) {
		// this nullability warning is wrong
		//noinspection ConstantValue
		if (value == null) {
			//noinspection unchecked
			return CompletedTask.COMPLETED_NULL;
		}
		return new CompletedTask<>(value);
	}

	static <V> Task<V> wrap(Future<V> future) {
		return new FutureAsTask<>(future);
	}

	static <V> Task<V> nonCancellable(Task<V> task) {
		return switch (task) {
			case Task.Cancellable<V> cancellable -> new NonCancellableTask<>(cancellable);
			default -> task;
		};
	}

	static <V> V joinAndCancelOnInterrupt(Cancellable<V> task) throws TaskException {
		try {
			return task.join();
		} catch (InterruptedException e) {
			task.cancel();
			throw new TaskException.Cancelled("Join interrupted", e);
		}
	}

	static <V> V joinAndCancelOnInterruptIfPossible(Task<V> task) throws TaskException {
		try {
			return task.join();
		} catch (InterruptedException e) {
			if (task instanceof Task.Cancellable<V> cancellable) {
				cancellable.cancel();
				throw new TaskException.Cancelled("Join interrupted", e);
			}
			throw new RuntimeException("Join interrupted, but task could not be cancelled", e);
		}
	}

	/**
	 * Spawns a new virtual thread that joins the {@code task} and does not attempt to catch exceptions. If one is
	 * thrown, the default uncaught exception handler for the VM will be invoked.
	 */
	static void uncatch(Task<?> task) {
		Thread.startVirtualThread(() -> {
			try {
				task.join();
			} catch (InterruptedException e) {
				throw new RuntimeException(e);
			}
		});
	}
}
